@@ -1,6 +1,5 @@
 import mujoco
 import glfw
-import sys
 from threading import Lock
 import numpy as np
 import time
@@ -8,15 +7,30 @@ import imageio
 
 
 class MujocoViewer:
-    def __init__(self, model, data):
+    def __init__(
+            self,
+            model,
+            data,
+            mode='window',
+            title="mujoco-python-viewer",
+            width=None,
+            height=None,
+            hide_menus=True):
         self.model = model
         self.data = data
+        self.render_mode = mode
+        if self.render_mode not in ['offscreen', 'window']:
+            raise NotImplementedError("Invalid mode. Only 'offscreen' and 'window' are supported.")
+
+        self.is_alive = True
 
         self._gui_lock = Lock()
         self._button_left_pressed = False
         self._button_right_pressed = False
         self._left_double_click_pressed = False
         self._right_double_click_pressed = False
+        self._last_left_click_time = None
+        self._last_right_click_time = None
         self._last_mouse_x = 0
         self._last_mouse_y = 0
         self._paused = False
@@ -35,29 +49,39 @@ class MujocoViewer:
         self._run_speed = 1.0
         self._loop_count = 0
         self._advance_by_one_step = False
-        self._hide_menu = False
+        self._hide_menus = hide_menus
 
         # glfw init
         glfw.init()
-        width, height = glfw.get_video_mode(glfw.get_primary_monitor()).size
+
+        if not width:
+            width, _ = glfw.get_video_mode(glfw.get_primary_monitor()).size
+
+        if not height:
+            _, height = glfw.get_video_mode(glfw.get_primary_monitor()).size
+
+        if self.render_mode=='offscreen':
+            glfw.window_hint(glfw.VISIBLE, 0)
+
         self.window = glfw.create_window(
-            width // 2, height // 2, "mujoco", None, None)
+            width, height, title, None, None)
         glfw.make_context_current(self.window)
         glfw.swap_interval(1)
 
         framebuffer_width, framebuffer_height = glfw.get_framebuffer_size(
             self.window)
-        window_width, _ = glfw.get_window_size(self.window)
-        self._scale = framebuffer_width * 1.0 / window_width
 
-        # set callbacks
-        glfw.set_cursor_pos_callback(self.window, self._cursor_pos_callback)
-        glfw.set_mouse_button_callback(
-            self.window, self._mouse_button_callback)
-        glfw.set_scroll_callback(self.window, self._scroll_callback)
-        glfw.set_key_callback(self.window, self._key_callback)
-        self._last_left_click_time = None
-        self._last_right_click_time = None
+        # install callbacks only for 'window' mode
+        if self.render_mode=='window':
+            window_width, _ = glfw.get_window_size(self.window)
+            self._scale = framebuffer_width * 1.0 / window_width
+
+            # set callbacks
+            glfw.set_cursor_pos_callback(self.window, self._cursor_pos_callback)
+            glfw.set_mouse_button_callback(
+                self.window, self._mouse_button_callback)
+            glfw.set_scroll_callback(self.window, self._scroll_callback)
+            glfw.set_key_callback(self.window, self._key_callback)
 
         # create options, camera, scene, context
         self.vopt = mujoco.MjvOption()
@@ -77,6 +101,8 @@ class MujocoViewer:
 
     def _key_callback(self, window, key, scancode, action, mods):
         if action != glfw.RELEASE:
+            if key == glfw.KEY_LEFT_ALT:
+                self._hide_menus = False
             return
         # Switch cameras
         elif key == glfw.KEY_TAB:
@@ -122,8 +148,10 @@ class MujocoViewer:
         elif key == glfw.KEY_E:
             self.vopt.frame = 1 - self.vopt.frame
         # Hide overlay menu
+        elif key == glfw.KEY_LEFT_ALT:
+            self._hide_menus = True
         elif key == glfw.KEY_H:
-            self._hide_menu = not self._hide_menu
+            self._hide_menus = not self._hide_menus
         # Make transparent
         elif key == glfw.KEY_R:
             self._transparent = not self._transparent
@@ -160,8 +188,8 @@ class MujocoViewer:
         if key == glfw.KEY_ESCAPE:
             print("Pressed ESC")
             print("Quitting.")
-            glfw.terminate()
-            sys.exit(0)
+            glfw.set_window_should_close(self.window, True)
+        return
 
     def _cursor_pos_callback(self, window, xpos, ypos):
         if not (self._button_left_pressed or self._button_right_pressed):
@@ -441,7 +469,7 @@ class MujocoViewer:
             topleft,
             "Referenc[e] frames",
             "On" if self.vopt.frame == 1 else "Off")
-        add_overlay(topleft, "[H]ide Menu", "")
+        add_overlay(topleft, "[H]ide Menus", "")
         if self._image_idx > 0:
             fname = self._image_path % (self._image_idx - 1)
             add_overlay(topleft, "Cap[t]ure frame", "Saved as %s" % fname)
@@ -466,18 +494,56 @@ class MujocoViewer:
         mujoco.mjv_applyPerturbPose(self.model, self.data, self.pert, 0)
         mujoco.mjv_applyPerturbForce(self.model, self.data, self.pert)
 
+    def read_pixels(self, camid=None):
+        if self.render_mode is 'window':
+            raise NotImplementedError(
+                "Use 'render()' in 'window' mode.")
+
+        if camid is not None:
+            if camid==-1:
+                self.cam.type = mujoco.mjtCamera.mjCAMERA_FREE
+            else:
+                self.cam.type = mujoco.mjtCamera.mjCAMERA_FIXED
+            self.cam.fixedcamid = camid
+
+        self.viewport.width, self.viewport.height = glfw.get_framebuffer_size(
+            self.window)
+        # update scene
+        mujoco.mjv_updateScene(
+            self.model,
+            self.data,
+            self.vopt,
+            self.pert,
+            self.cam,
+            mujoco.mjtCatBit.mjCAT_ALL.value,
+            self.scn)
+        # render
+        mujoco.mjr_render(self.viewport, self.scn, self.ctx)
+
+        img = np.zeros(
+            (glfw.get_framebuffer_size(
+                self.window)[1], glfw.get_framebuffer_size(
+                self.window)[0], 3), dtype=np.uint8)
+        mujoco.mjr_readPixels(img, None, self.viewport, self.ctx)
+        return np.flipud(img)
+
     def render(self):
+        if self.render_mode is 'offscreen':
+            raise NotImplementedError(
+                "Use 'read_pixels()' for 'offscreen' mode.")
+        if not self.is_alive:
+            raise Exception(
+                "GLFW window does not exist but you tried to render.")
+        if glfw.window_should_close(self.window):
+            self.close()
+            return
+
         # mjv_updateScene, mjr_render, mjr_overlay
         def update():
             # fill overlay items
             self._create_overlay()
 
             render_start = time.time()
-            if self.window is None:
-                return
-            elif glfw.window_should_close(self.window):
-                glfw.terminate()
-                sys.exit(0)
             self.viewport.width, self.viewport.height = glfw.get_framebuffer_size(
                 self.window)
             with self._gui_lock:
@@ -496,15 +562,19 @@ class MujocoViewer:
                 # render
                 mujoco.mjr_render(self.viewport, self.scn, self.ctx)
                 # overlay items
-                if not self._hide_menu:
-                    for gridpos, [t1, t2] in self._overlay.items():
-                        mujoco.mjr_overlay(
-                            mujoco.mjtFontScale.mjFONTSCALE_150,
-                            gridpos,
-                            self.viewport,
-                            t1,
-                            t2,
-                            self.ctx)
+                for gridpos, [t1, t2] in self._overlay.items():
+                    menu_positions = [mujoco.mjtGridPos.mjGRID_TOPLEFT,
+                                      mujoco.mjtGridPos.mjGRID_BOTTOMLEFT]
+                    if gridpos in menu_positions and self._hide_menus:
+                        continue
+
+                    mujoco.mjr_overlay(
+                        mujoco.mjtFontScale.mjFONTSCALE_150,
+                        gridpos,
+                        self.viewport,
+                        t1,
+                        t2,
+                        self.ctx)
                 glfw.swap_buffers(self.window)
             glfw.poll_events()
             self._time_per_render = 0.9 * self._time_per_render + \
@@ -516,6 +586,9 @@ class MujocoViewer:
         if self._paused:
             while self._paused:
                 update()
+                if glfw.window_should_close(self.window):
+                    self.close()
+                    break
                 if self._advance_by_one_step:
                     self._advance_by_one_step = False
                     break
@@ -535,5 +608,6 @@ class MujocoViewer:
         self.apply_perturbations()
 
     def close(self):
+        self.is_alive = False
         glfw.terminate()
         self.ctx.free()
